//


#include "predictor.h"
#include "resource.h"


TCHAR szDescription[] = TEXT("predictor v0.0a");
TCHAR szName[]        = TEXT("predictor");




const int cyb = int(0.114*219/255*65536+0.5);
const int cyg = int(0.587*219/255*65536+0.5);
const int cyr = int(0.299*219/255*65536+0.5);


#define VERSION         0x00000000      // 0.0


   /*
    * pred.c -- Test program for Dave Rand's rendition of the
    * predictor algorithm
    * Updated by: iand@labtam.labtam.oz.au (Ian Donaldson)
    * Updated by: Carsten Bormann <cabo@cs.tu-berlin.de>
    * Original  : Dave Rand <dlr@bungi.com>/<dave_rand@novell.com>
    */

   /* The following hash code is the heart of the algorithm:
    * It builds a sliding hash sum of the previous 3-and-a-bit
    * characters which will be used to index the guess table.
    * A better hash function would result in additional compression,
    * at the expense of time.
    */
   #define HASH(x) Hash = (Hash << 4) ^ (x)

   static unsigned short int Hash;
   static unsigned char GuessTable[65536];


   static inline int compress(unsigned char *source, unsigned char *dest, int len)
   {
    unsigned char *orgdest;

    orgdest = dest;

/* 
 ESI = SOURCE
 EDI = DEST
 ECX = LEN
 EBX = HASH
 AL = FLAGS
 AH = BITMASK
 EBP = FLAGDEST
 
*/
__asm  pusha
__asm  push esi
__asm  push edi

__asm  sub ebx, ebx
__asm  mov esi, source
__asm  mov edi, dest
__asm  mov ecx, len
__asm  mov bx, Hash
__asm  push ebp

__asm main_loop:
__asm  mov dl, [esi]
__asm  mov ebp, edi
__asm  inc edi
__asm  sub al, al
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_1
__asm  or al,1
__asm  jmp end_pred_1
__asm no_equal_1:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_1:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop

__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_2
__asm  or al,2
__asm  jmp end_pred_2
__asm no_equal_2:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_2:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop

__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_4
__asm  or al,4
__asm  jmp end_pred_4
__asm no_equal_4:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_4:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop


__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_8
__asm  or al,8
__asm  jmp end_pred_8
__asm no_equal_8:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_8:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop


__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_16
__asm  or al,16
__asm  jmp end_pred_16
__asm no_equal_16:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_16:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop


__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_32
__asm  or al,32
__asm  jmp end_pred_32
__asm no_equal_32:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_32:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop


__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_64
__asm  or al,64
__asm  jmp end_pred_64
__asm no_equal_64:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_64:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jz  end_main_loop

__asm  mov dl, [esi]
__asm  cmp dl, GuessTable[ebx]
__asm  jne no_equal_128
__asm  or al,128
__asm  jmp end_pred_128
__asm no_equal_128:
__asm  mov [edi], dl
__asm  mov GuessTable[ebx], dl
__asm  inc edi
__asm end_pred_128:
__asm  sub edx,edx
__asm  shl bx, 4
__asm  mov dl, [esi]
__asm  inc esi
__asm  xor bx, dx
__asm  dec ecx
__asm  mov [ebp], al
__asm  jnz main_loop


__asm end_main_loop:
__asm  pop ebp

__asm  mov dest, edi
__asm  mov Hash, bx
__asm  pop edi
__asm  pop esi
__asm  popa

/*
       register bitmask;
       unsigned char *flagdest, flags, *orgdest;

       orgdest = dest;
       for(;;)
	   {
		   if(!len)
			break;
           flagdest = dest++; flags = 0; // All guess wrong initially
           bitmask=1;
 
               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=2;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output 
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char 
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=4;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=8;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output 
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=16;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char 
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=32;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char 
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=64;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output 
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char 
               }//else
               HASH(*source++);len--;
			   // Prediction - END

		   if(!len)
		   {
  	        *flagdest = flags;
			break;
		   }//if
           bitmask=128;

               // Prediction - BEGIN
               if (GuessTable[Hash] == *source) 
			   {
                   flags |= bitmask; // Guess was right - don't output
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;
                   *dest++ = *source; // Guess wrong, output char
               }//else
               HASH(*source++);len--;
			   // Prediction - END
			   

	    *flagdest = flags;
       }//for
*/

       return(dest - orgdest);
   }//compress

   static inline int decompress(unsigned char *source, unsigned char *dest, int *lenp, int final)
   {
       int i, bitmask;
       unsigned char flags, *orgdest;
       int len = *lenp;
       orgdest = dest;
       while (len >= 9) 
	   {
           flags = *source++;
           for (i=0, bitmask = 1; i < 8; i++, bitmask <<= 1) 
		   {
               if (flags & bitmask) 
			   {
                   *dest = GuessTable[Hash];       /* Guess correct */
               }//if 
			   else 
			   {
                   GuessTable[Hash] = *source;     /* Guess wrong */
                   *dest = *source++;          /* Read from source */
                   len--;
               }//else
               HASH(*dest++);
           }//for
           len--;
       }//while
       while (final && len) 
	   {
           flags = *source++;
           len--;
           for (i=0, bitmask = 1; i < 8; i++, bitmask <<= 1) 
		   {
               if (flags & bitmask) 
			   {
                   *dest = GuessTable[Hash];       /* Guess correct */
               }//if 
			   else 
			   {
                   if (!len)
                       break;  /* we seem to be really done -- cabo */
                   GuessTable[Hash] = *source;     /* Guess wrong */
                   *dest = *source++;          /* Read from source */
                   len--;
               }//else
               HASH(*dest++);
           }//for
       }//while
       *lenp = len;
       return(dest - orgdest);
   }




/********************************************************************
********************************************************************/
CodecInst::CodecInst() 
{ 
  yuy2_buffer=NULL;
  yuy2_buffer_old[0]=NULL;
  yuy2_buffer_old[1]=NULL;
  FrameCount=0;
  icinfo_sel=3;

}//CodecInst

/********************************************************************
********************************************************************/

void Msg(const char fmt[], ...) {
  static int debug = GetPrivateProfileInt("debug", "log", 0, "predictor.ini");
  if (!debug) return;

  DWORD written;
  char buf[2000];
  va_list val;
  
  va_start(val, fmt);
  wvsprintf(buf, fmt, val);

  const COORD _80x50 = {80,50};
  static BOOL startup = (AllocConsole(), SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), _80x50));
  WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), buf, lstrlen(buf), &written, 0);
}


int AppFlags() {
  static int flags = -1;
  if (flags < 0) {
    flags = 0;
    TCHAR apppath[MAX_PATH];
    if (GetModuleFileName(NULL, apppath, MAX_PATH)) {
      TCHAR* appname = strrchr(apppath, '\\');
      appname = appname ? appname+1 : apppath;
      Msg("App name is %s; ", appname);
      if (!lstrcmpi(appname, TEXT("premiere.exe")))
        flags = 1;
      if (!lstrcmpi(appname, TEXT("veditor.exe")))
        flags = 1;
      if (!lstrcmpi(appname, TEXT("afterfx.exe")))
        flags = 2;
      Msg("flags=%d\n", flags);
    }
  }
  return flags;
}


/********************************************************************
********************************************************************/

CodecInst* Open(ICOPEN* icinfo) {
  if (icinfo && icinfo->fccType != ICTYPE_VIDEO)
      return NULL;

  CodecInst* pinst = new CodecInst();

  if (icinfo) icinfo->dwError = pinst ? ICERR_OK : ICERR_MEMORY;

  return pinst;
}

DWORD Close(CodecInst* pinst) {
//    delete pinst;       // this caused problems when deleting at app close time
    return 1;
}

/********************************************************************
********************************************************************/

BOOL CodecInst::QueryAbout() { return TRUE; }

static BOOL CALLBACK AboutDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_COMMAND) {
    switch (LOWORD(wParam)) {
    case IDOK:
      EndDialog(hwndDlg, 0);
      break;
    case IDC_HOMEPAGE:
      ShellExecute(NULL, NULL, "http://www.amacias.tk/", NULL, NULL, SW_SHOW);
      break;
    case IDC_EMAIL:
      ShellExecute(NULL, NULL, "mailto:amp_21004@yahoo.es", NULL, NULL, SW_SHOW);
      break;
    }
  }
  return FALSE;
}

DWORD CodecInst::About(HWND hwnd) {
  DialogBox(hmodulepredictor, MAKEINTRESOURCE(IDD_ABOUT), hwnd, AboutDialogProc);
  return ICERR_OK;
}

static BOOL CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_INITDIALOG) 
  {
    CheckDlgButton(hwndDlg, IDC_RGBOUTPUT,
      GetPrivateProfileInt("debug", "rgboutput", 0, "predictor.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_LOG,
      GetPrivateProfileInt("debug", "log", 0, "predictor.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_MOTION,
      GetPrivateProfileInt("debug", "motion", 0, "predictor.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_ENSURE,
      GetPrivateProfileInt("debug", "ensure", 0, "predictor.ini") ? BST_CHECKED : BST_UNCHECKED);
  } 
  else 
  if (uMsg == WM_COMMAND) 
  {
    switch (LOWORD(wParam)) 
	{
    case IDOK:
      WritePrivateProfileString("debug", "rgboutput",
        (IsDlgButtonChecked(hwndDlg, IDC_RGBOUTPUT) == BST_CHECKED) ? "1" : NULL, "predictor.ini");
      WritePrivateProfileString("debug", "log",
        (IsDlgButtonChecked(hwndDlg, IDC_LOG) == BST_CHECKED) ? "1" : "0", "predictor.ini");
      WritePrivateProfileString("debug", "motion",
        (IsDlgButtonChecked(hwndDlg, IDC_MOTION) == BST_CHECKED) ? "1" : "0", "predictor.ini");
      WritePrivateProfileString("debug", "ensure",
        (IsDlgButtonChecked(hwndDlg, IDC_ENSURE) == BST_CHECKED) ? "1" : "0", "predictor.ini");
    case IDCANCEL:
      EndDialog(hwndDlg, 0);
      break;
    default:
      return AboutDialogProc(hwndDlg, uMsg, wParam, lParam);    // handle email and home-page buttons
    }
  }

  return FALSE;
}

BOOL CodecInst::QueryConfigure() { return TRUE; }

DWORD CodecInst::Configure(HWND hwnd) {
  DialogBox(hmodulepredictor, MAKEINTRESOURCE(IDD_CONFIGURE), hwnd, ConfigureDialogProc);
  return ICERR_OK;
}


// we have no state information which needs to be stored

DWORD CodecInst::GetState(LPVOID pv, DWORD dwSize) { return 0; }

DWORD CodecInst::SetState(LPVOID pv, DWORD dwSize) { return 0; }


DWORD CodecInst::GetInfo(ICINFO* icinfo, DWORD dwSize) {
  if (icinfo == NULL)
    return sizeof(ICINFO);

  if (dwSize < sizeof(ICINFO))
    return 0;

  icinfo->dwSize            = sizeof(ICINFO);
  icinfo->fccType           = ICTYPE_VIDEO;
  icinfo->fccHandler        = FOURCC_HFYU;
  icinfo->dwFlags           = 0;

  icinfo->dwVersion         = VERSION;
  icinfo->dwVersionICM      = ICVERSION;
  MultiByteToWideChar(CP_ACP, 0, szDescription, -1, icinfo->szDescription, sizeof(icinfo->szDescription)/sizeof(WCHAR));
  MultiByteToWideChar(CP_ACP, 0, szName, -1, icinfo->szName, sizeof(icinfo->szName)/sizeof(WCHAR));

  return sizeof(ICINFO);
}


/********************************************************************
********************************************************************/


struct PrintBitmapType {
  char s[32];
  PrintBitmapType(LPBITMAPINFOHEADER lpbi) {
    if (!lpbi)
      strcpy(s,  "(null)");
    else {
      *(DWORD*)s = lpbi->biCompression;
      s[4] = 0;
      if (!isalnum(s[0]) || !isalnum(s[1]) || !isalnum(s[2]) || !isalnum(s[3]))
        wsprintfA(s, "%x", lpbi->biCompression);
      wsprintfA(strchr(s, 0), ", bits = %d", lpbi->biBitCount);
    }
  }
};


// fast clipping of values to unsigned char range

static unsigned char clip[896];

static void InitClip() {
  memset(clip, 0, 320);
  for (int i=0; i<256; ++i) clip[i+320] = i;
  memset(clip+320+256, 255, 320);
}

/*
static inline unsigned char Clip(int x)
  { return clip[320 + ((x+0x8000) >> 16)]; }
*/

#define Clip(x)  clip[320 + (((x)+0x8000) >> 16)]


/********************************************************************
********************************************************************/


// 0=unknown, -1=compressed YUY2, -2=compressed RGB, -3=compressed RGBA, 1=YUY2, 2=UYVY, 3=RGB 24-bit, 4=RGB 32-bit
static int GetBitmapType(LPBITMAPINFOHEADER lpbi) {
  if (!lpbi)
    return 0;
  if (lpbi->biCompression == FOURCC_VYUY || lpbi->biCompression == FOURCC_YUY2)
    return 1;
  if (lpbi->biCompression == FOURCC_UYVY)
    return 2;
  if (lpbi->biCompression == 0 || lpbi->biCompression == ' BID')
    return (lpbi->biBitCount == 24) ? 3 : (lpbi->biBitCount == 32) ? 4 : 0;
  if (lpbi->biCompression == FOURCC_HFYU)
    return (lpbi->biBitCount == 16) ? -1 : (lpbi->biBitCount == 24) ? -2 : (lpbi->biBitCount == 32) ? -3 : 0;
  return 0;
}


static bool CanCompress(LPBITMAPINFOHEADER lpbiIn) {
  int intype = GetBitmapType(lpbiIn);
  return (intype == 1 || intype == 2 || intype == 3 || (intype == 4 && (AppFlags()&2)));
}


static bool CanCompress(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (!lpbiOut) return CanCompress(lpbiIn);
  int outtype = GetBitmapType(lpbiOut);
  switch (GetBitmapType(lpbiIn)) {
    case 1: case 2:
      return (outtype == -1);
    case 3:
      return (outtype == 1 || outtype == -1 || outtype == -2);
    case 4:
      return (outtype == -3 && (AppFlags() & 2));
    default:
      return false;
  }
}


/********************************************************************
********************************************************************/


DWORD CodecInst::CompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("CompressQuery: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanCompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


DWORD CodecInst::CompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (lpbiOut == 0)
    return sizeof(BITMAPINFOHEADER);

  if (!CanCompress(lpbiIn))
    return ICERR_BADFORMAT;

  bool compress_as_yuy2 = !!GetPrivateProfileInt("general", "compress_as_yuy2", false, "predictor.ini");

  *lpbiOut = *lpbiIn;
  lpbiOut->biCompression = FOURCC_HFYU;
  int intype = GetBitmapType(lpbiIn);
  lpbiOut->biBitCount = (intype==4) ? 32 : (intype==1 || intype==2 || compress_as_yuy2) ? 16 : 24;
  return ICERR_OK;
}


DWORD CodecInst::CompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
//  total_bpp = 0;
//  num_frames = 0;
  Msg("CompressBegin: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));

  CompressEnd();  // free resources if necessary

  enable_motion = !!GetPrivateProfileInt("debug", "motion", false, "predictor.ini");

  enable_ensure = !!GetPrivateProfileInt("debug", "ensure", false, "predictor.ini");
  if(enable_ensure)
	  enable_motion=1;

  // allocate buffer if compressing RGB->YUY2->HFYU
  if (GetBitmapType(lpbiIn) == 3)
    yuy2_buffer = new unsigned char[(lpbiIn->biWidth * lpbiIn->biHeight * 2)*9/8];
  if(enable_motion)
  {
    yuy2_buffer_old[0] = new unsigned char[(lpbiIn->biWidth * lpbiIn->biHeight * 2)*9/8];
    yuy2_buffer_old[1] = new unsigned char[(lpbiIn->biWidth * lpbiIn->biHeight * 2)*9/8];
  }//if

  icinfo_sel=0;
  FrameCount=64;

  InitClip();

  return CanCompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}

DWORD CodecInst::CompressGetSize(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  // Assume 24 bpp worst-case for YUY2 input, 40 bpp for RGB.
  // The actual worst case is 43/51 bpp currently, but this is exceedingly improbable
  // (probably impossible with real captured video)
  return lpbiIn->biWidth * lpbiIn->biHeight * ((GetBitmapType(lpbiIn) <= 2) ? 3 : 5);
}





DWORD CodecInst::Compress(ICCOMPRESS* icinfo, DWORD dwSize) 
{
  unsigned char* input = (unsigned char*)icinfo->lpInput;
  unsigned char* end;
  unsigned SizeImage=(icinfo->lpbiInput->biWidth * icinfo->lpbiInput->biHeight)<<1;
  unsigned char* icinfo_ld=(unsigned char*)icinfo->lpPrev;
  unsigned i,j,k,l,m,n,o,p;

  if (GetBitmapType(icinfo->lpbiInput) == 3) 
  {
   ConvertRGB24toYUY2(input, yuy2_buffer, icinfo->lpbiInput->biWidth, icinfo->lpbiInput->biHeight);
   input = yuy2_buffer;
  }//if

  end = input + SizeImage;
  unsigned char* output = (unsigned char*)icinfo->lpOutput;


 if(enable_motion)
 {
      unsigned EndLoop=SizeImage-8;

	  memcpy(yuy2_buffer_old[icinfo_sel],input,SizeImage);
      icinfo_sel=(icinfo_sel + 1) & 1;
      icinfo_ld=yuy2_buffer_old[icinfo_sel];

      if((FrameCount==64) || (icinfo->dwFlags == ICCOMPRESS_KEYFRAME))
	  {
	   //insert a key frame every 64 frames
	   FrameCount=0;
       *icinfo->lpdwFlags = AVIIF_KEYFRAME;
	  }
	  else
	  {
       j=1;
	   k=2;
	   l=3;
	   m=4;
	   n=5;
	   o=6;
	   p=7;
	   // Substract the last frame from the actual
	   for(i=0; i<EndLoop;)
	   {
		  input[i] -= icinfo_ld[i];
		  i+=8;
		  input[j] -= icinfo_ld[j];
		  j+=8;
		  input[k] -= icinfo_ld[k];
		  k+=8;
		  input[l] -= icinfo_ld[l];
		  l+=8;
		  input[m] -= icinfo_ld[m];
		  m+=8;
		  input[n] -= icinfo_ld[n];
		  n+=8;
		  input[o] -= icinfo_ld[o];
		  o+=8;
		  input[p] -= icinfo_ld[p];
		  p+=8;
	   }//for
      }//else
      FrameCount++;
 }//if
 else
 {
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;
 }//else


  memset(GuessTable,0,65536);
  Hash=0;
  icinfo->lpbiOutput->biSizeImage = compress(input,output,SizeImage);

  if(enable_ensure && !(*icinfo->lpdwFlags == AVIIF_KEYFRAME))
  {
      memset(GuessTable,0,65536);
      Hash=0;
	  unsigned Size=compress(yuy2_buffer_old[(icinfo_sel+1)&1],input,SizeImage);
	  if(Size<icinfo->lpbiOutput->biSizeImage)
	  {
        icinfo->lpbiOutput->biSizeImage=Size;
		memcpy(output,input,Size);
        *icinfo->lpdwFlags = AVIIF_KEYFRAME;
		FrameCount=0;
	  }//if
  }//if


  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;


  return ICERR_OK;


}//Compress


inline void CodecInst::ConvertRGB24toYUY2(unsigned char* src, unsigned char* dst, int width, int height)
{
 unsigned char* rgb;
 unsigned char* yuv;
 int width3=width*3;
 int width2=width<<1;
 int y1,y2,scaled_y,b_y,r_y;
 


  for (int row = 0; row < height; ++row) 
  {
    rgb = src + width3 * (height-1-row);
    yuv = dst + width2 * row;
    for (int col = 0; col < width; col += 8) 
	{
      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

    }
  }
}//ConvertRGB24toYUY2


DWORD CodecInst::CompressEnd() 
{
  if (yuy2_buffer) 
  {
    delete[] yuy2_buffer;
    yuy2_buffer = 0;
  }//if

  if (yuy2_buffer_old[0]) 
  {
    delete[] yuy2_buffer_old[0];
    yuy2_buffer_old[0] = 0;
    delete[] yuy2_buffer_old[1];
    yuy2_buffer_old[1] = 0;
  }//if
  

  return ICERR_OK;
}//CompressEnd


/********************************************************************
********************************************************************/

static bool CanDecompress(LPBITMAPINFOHEADER lpbiIn) {
  int intype = GetBitmapType(lpbiIn);
  return (intype < 0 || intype == 1 || intype == 2);
}


static bool CanDecompress(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (!lpbiOut)
    return CanDecompress(lpbiIn);

  // must be 1:1 (no stretching)
  if (lpbiOut && (lpbiOut->biWidth != lpbiIn->biWidth || lpbiOut->biHeight != lpbiIn->biHeight))
    return false;

  int outtype = GetBitmapType(lpbiOut);
  switch (GetBitmapType(lpbiIn)) {
    case -1:
      // YUY2, RGB-24, RGB-32 output for compressed YUY2
      return (outtype == 1 || outtype == 3 || outtype == 4);
    case -2: case 1: case 2:
      // RGB-24, RGB-32 output only for YUY2/UYVY and compressed RGB
      return (outtype == 3 || outtype == 4);
    case -3:
      // RGB-32 output for compressed RGBA
      return (outtype == 4);
    default:
      return false;
  }
}

/********************************************************************
********************************************************************/


DWORD CodecInst::DecompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("DecompressQuery: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanDecompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


// This function should return "the output format which preserves the most
// information."  However, I now provide the option to return RGB format
// instead, since some programs treat the default format as the ONLY format.

DWORD CodecInst::DecompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {

  // if lpbiOut == NULL, then return the size required to hold an output format struct
  if (lpbiOut == NULL)
    return lpbiIn->biSize;

  if (!CanDecompress(lpbiIn))
    return ICERR_BADFORMAT;

  // get default rgboutput value and allow the user to override
  bool rgboutput = AppFlags() & 1;
  rgboutput = !!GetPrivateProfileInt("debug", "rgboutput", rgboutput, "predictor.ini");

  memcpy(lpbiOut, lpbiIn, lpbiIn->biSize);

  int intype = GetBitmapType(lpbiIn);
  if (intype == -3) {
    lpbiOut->biBitCount = 32;   // "RGBA" (though not really)
    lpbiOut->biCompression = 0;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 4;
  } else if (rgboutput || intype == -2 || intype == 1 || intype == 2) {
    lpbiOut->biBitCount = 24;   // RGB
    lpbiOut->biCompression = 0;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 3;
  } else {
    lpbiOut->biBitCount = 16;       // YUY2
    lpbiOut->biCompression = FOURCC_YUY2;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 2;
  }

  return ICERR_OK;
}

DWORD CodecInst::DecompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("DecompressBegin: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));

  DecompressEnd();  // free resources if necessary

  // allocate buffer if decompressing HFYU->YUY2->RGB
  yuy2_buffer_old[0] = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  yuy2_buffer_old[1] = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  icinfo_sel=0;

  InitClip();

  return CanDecompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}




static inline void YUV2RGB(int y, int u, int v, BYTE* out) {
  const int crv = int(1.596*65536+0.5);
  const int cgv = int(0.813*65536+0.5);
  const int cgu = int(0.391*65536+0.5);
  const int cbu = int(2.018*65536+0.5);

  int scaled_y = (y - 16) * int((255.0/219.0)*65536+0.5);

  out[0] = Clip(scaled_y + (u-128) * cbu); // blue
  out[1] = Clip(scaled_y - (u-128) * cgu - (v-128) * cgv); // green
  out[2] = Clip(scaled_y + (v-128) * crv); // red
}


// I originally tried to do this with a template function, but it produced
// incorrect code.  So, since you insist...

#define ConvertYUVtoRGB(uyvy,rgb_bytes) \
  const int width = lpbiOutput->biWidth; \
  const int height = lpbiOutput->biHeight; \
  lpbiOutput->biSizeImage = width * height * rgb_bytes; \
\
  unsigned char* const yuv_base = (unsigned char*)src; \
  unsigned char* const rgb_base = (unsigned char*)dst; \
  for (int row = 0; row < height; ++row) { \
    unsigned char* yuv = yuv_base + width * 2 * (height-1-row); \
    unsigned char* rgb = rgb_base + width * rgb_bytes * row; \
    YUV2RGB(yuv[0+uyvy], yuv[1-uyvy], yuv[3-uyvy], rgb); \
    YUV2RGB(yuv[2+uyvy], yuv[1-uyvy], yuv[3-uyvy], rgb+rgb_bytes); \
    if (rgb_bytes==4) \
      rgb[7] = rgb[3] = 0; \
    for (int col = 2; col < width; col += 2) { \
      yuv += 4; \
      rgb += rgb_bytes*2; \
      /* interpolate chroma values */ \
      YUV2RGB(yuv[0+uyvy], (int(yuv[1-uyvy])+int(yuv[-3-uyvy]))/2, (int(yuv[3-uyvy])+int(yuv[-1-uyvy]))/2, rgb); \
      YUV2RGB(yuv[2+uyvy], yuv[1-uyvy], yuv[3-uyvy], rgb+rgb_bytes); \
      if (rgb_bytes==4) \
        rgb[7] = rgb[3] = 0; \
    } \
  } \
  return ICERR_OK


static DWORD ConvertYUY2toRGB24(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(false,3); }
static DWORD ConvertYUY2toRGB32(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(false,4); }
static DWORD ConvertUYVYtoRGB24(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(true,3); }
static DWORD ConvertUYVYtoRGB32(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(true,4); }


DWORD CodecInst::Decompress(ICDECOMPRESS* icinfo, DWORD dwSize) 
{

  int width = icinfo->lpbiInput->biWidth;
  int height = icinfo->lpbiInput->biHeight;
  bool decompress_to_rgb = (icinfo->lpbiOutput->biBitCount > 16);
  unsigned char* icinfo_old2;

  unsigned char* in = (unsigned char*)icinfo->lpInput;
  unsigned char* out=yuy2_buffer_old[icinfo_sel];


  int SizeImage=icinfo->lpbiInput->biSizeImage;
  int FrameSize=(width*height)<<1;
  int NewSize=0;

  memset(GuessTable,0,65536);
  Hash=0;


  decompress(in, out, &SizeImage, 1);


      
	  icinfo_sel = (icinfo_sel + 1) & 1;
      SizeImage=FrameSize;
	  icinfo_old2=yuy2_buffer_old[icinfo_sel];

	  // Add the last frame to the actual if it is not a key frame
	  if(icinfo->dwFlags & ICDECOMPRESS_NOTKEYFRAME)
	  {
	   for(int i=0; i<SizeImage; i++)
	 	  out[i] += icinfo_old2[i];
	  }

	  




  if (decompress_to_rgb) {
    if (icinfo->lpbiOutput->biBitCount == 32)
      ConvertYUY2toRGB32(icinfo->lpbiOutput, out, icinfo->lpOutput);
    else
      ConvertYUY2toRGB24(icinfo->lpbiOutput, out, icinfo->lpOutput);
  }

  return ICERR_OK;
}



// palette-mapped output only
DWORD CodecInst::DecompressGetPalette(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  return ICERR_BADFORMAT;
}


DWORD CodecInst::DecompressEnd() 
{
  if (yuy2_buffer) 
  {
    delete[] yuy2_buffer;
    yuy2_buffer = 0;
  }//if

  if (yuy2_buffer_old[0]) 
  {
    delete[] yuy2_buffer_old[0];
    yuy2_buffer_old[0] = 0;
    delete[] yuy2_buffer_old[1];
    yuy2_buffer_old[1] = 0;
  }//if


 return ICERR_OK;
} // DecompressEnd

